/* SHA512
 * Daniel Beer <dlbeer@gmail.com>, 22 Apr 2014
 *
 * This file is in the public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "sha512.h"

struct test_vector {
	const char  *text;
	uint8_t     hash[SHA512_HASH_SIZE];
};

static const struct test_vector test_vectors[] = {
	{
		.text = "",
		.hash = {
			0xcf, 0x83, 0xe1, 0x35, 0x7e, 0xef, 0xb8, 0xbd,
			0xf1, 0x54, 0x28, 0x50, 0xd6, 0x6d, 0x80, 0x07,
			0xd6, 0x20, 0xe4, 0x05, 0x0b, 0x57, 0x15, 0xdc,
			0x83, 0xf4, 0xa9, 0x21, 0xd3, 0x6c, 0xe9, 0xce,
			0x47, 0xd0, 0xd1, 0x3c, 0x5d, 0x85, 0xf2, 0xb0,
			0xff, 0x83, 0x18, 0xd2, 0x87, 0x7e, 0xec, 0x2f,
			0x63, 0xb9, 0x31, 0xbd, 0x47, 0x41, 0x7a, 0x81,
			0xa5, 0x38, 0x32, 0x7a, 0xf9, 0x27, 0xda, 0x3e,
		}
	},
	{
		.text = "Hello world",
		.hash = {
			0xb7, 0xf7, 0x83, 0xba, 0xed, 0x82, 0x97, 0xf0,
			0xdb, 0x91, 0x74, 0x62, 0x18, 0x4f, 0xf4, 0xf0,
			0x8e, 0x69, 0xc2, 0xd5, 0xe5, 0xf7, 0x9a, 0x94,
			0x26, 0x00, 0xf9, 0x72, 0x5f, 0x58, 0xce, 0x1f,
			0x29, 0xc1, 0x81, 0x39, 0xbf, 0x80, 0xb0, 0x6c,
			0x0f, 0xff, 0x2b, 0xdd, 0x34, 0x73, 0x84, 0x52,
			0xec, 0xf4, 0x0c, 0x48, 0x8c, 0x22, 0xa7, 0xe3,
			0xd8, 0x0c, 0xdf, 0x6f, 0x9c, 0x1c, 0x0d, 0x47,
		}
	},
	{
		.text =
"It was a bright cold day in April, and the clocks were striking "
"thirteen. Winston Smith, his chin nuzzled into his breast in an effort "
"to escape the vile wind, slipped quickly through the glass doors of "
"Victory Mansions, though not quickly enough to prevent a swirl of gritty "
"dust from entering along with him.\n",
		.hash = {
			0x8d, 0xd9, 0x4f, 0x6b, 0x83, 0x2f, 0x96, 0x7c,
			0x99, 0x8b, 0x85, 0xd4, 0x27, 0xe8, 0x4f, 0x37,
			0x4d, 0xdd, 0x98, 0x4a, 0x98, 0x9d, 0x84, 0xda,
			0x53, 0x1f, 0x0e, 0x81, 0xc9, 0x24, 0xd7, 0x4c,
			0xe4, 0x50, 0x57, 0xa9, 0xc5, 0x56, 0xbf, 0x7d,
			0x14, 0x05, 0x3e, 0x08, 0x60, 0xd3, 0x1d, 0xa8,
			0x31, 0x24, 0xa3, 0x4d, 0x02, 0xf5, 0x60, 0xfe,
			0x55, 0xf6, 0xc9, 0x49, 0x3f, 0xfd, 0xa9, 0xea,
		}
	},
	{
		.text = /* 124 bytes: will require two final blocks */
"It was a bright cold day in April, and the clocks were striking "
"thirteen. Winston Smith, his chin nuzzled into his breast in",
		.hash = {
			0x11, 0x57, 0xb8, 0x74, 0xfa, 0xa2, 0x00, 0x42,
			0x8b, 0x18, 0x5e, 0x6a, 0xf4, 0x7d, 0x91, 0x17,
			0x44, 0x6c, 0xcf, 0x3e, 0xac, 0x43, 0x13, 0xb5,
			0x86, 0xe5, 0xe7, 0x3a, 0x90, 0x19, 0xe6, 0x48,
			0x4c, 0x16, 0x2b, 0xe4, 0xec, 0xbe, 0x81, 0x40,
			0x62, 0xf4, 0x0b, 0x1d, 0xd6, 0x72, 0xbc, 0xf7,
			0x69, 0x76, 0x0d, 0x21, 0x68, 0x22, 0x84, 0x26,
			0x28, 0xc2, 0xcd, 0x08, 0x79, 0x3c, 0xa3, 0x75,
		}
	},
};

#define NUM_VECTORS (sizeof(test_vectors) / sizeof(test_vectors[0]))

static void test(const struct test_vector *t)
{
	const unsigned int len = strlen(t->text);
	struct sha512_state s;
	uint8_t hash[SHA512_HASH_SIZE];
	unsigned int i;

	printf("Message (%d bytes): %s\n", len, t->text);

	/* Test hashing */
	sha512_init(&s);
	for (i = 0; i + SHA512_BLOCK_SIZE <= len; i += SHA512_BLOCK_SIZE)
		sha512_block(&s, (const uint8_t *)t->text + i);
	sha512_final(&s, (const uint8_t *)t->text + i, len);
	sha512_get(&s, hash, 0, SHA512_HASH_SIZE);

	printf("Hash: ");
	for (i = 0; i < sizeof(hash); i++)
		printf("%02x", hash[i]);
	printf("\n");

	assert(!memcmp(hash, t->hash, SHA512_HASH_SIZE));

	/* Test slicing */
	for (i = 0; i < 100; i++) {
		const int offset = random() % SHA512_HASH_SIZE;
		const int rlen = random() % (SHA512_HASH_SIZE - offset + 1);
		uint8_t slice[SHA512_HASH_SIZE];

		sha512_get(&s, slice, offset, rlen);
		assert(!memcmp(hash + offset, slice, rlen));
	}
}

int main(void)
{
	unsigned int i;

	srandom(0);

	for (i = 0; i < NUM_VECTORS; i++) {
		test(&test_vectors[i]);
		printf("\n");
	}

	return 0;
}
